import 'dart:async';

import 'package:flutter/material.dart';
import 'package:rxdart/subjects.dart';

class Query<T> {
  final BehaviorSubject<QueryState<T>> stateStream =
      BehaviorSubject.seeded(QueryState<T>(), sync: true);

  QueryConfig config;

  Query(this.config);

  Future refetch() {
    if (config.isInfinite) {
      final data = stateStream.value?.data;

      if (data is List) {
        // Clear the latest
        stateStream.value?.snapshot = AsyncSnapshot.waiting();
        stateStream.add(stateStream.value);

        var future = fetch();
        data.forEach((d) {
          future = future.then((_) => fetch());
        });
        return future;
      }
    } else {
      return fetch();
    }
  }

  Future fetch() {
    Future queryFn;

    if (stateStream.value.data == null) {
      stateStream.add(QueryState(
        snapshot: AsyncSnapshot<T>.waiting(),
      ));
    }

    final current = stateStream.value;

    if (config.isInfinite) {
      var latestPage = current?.data;
      // TODO: Make this less confusing
      var pages = [];
      if (latestPage is List) {
        pages = (latestPage as List);
        latestPage = pages.last;
      }

      final cursor = config.getFetchMore(latestPage, pages);

      if (cursor != false && cursor != null) {
        // TODO: Find a better way to do this than just mutating state
        if (current != null) {
          current.isFetchingMore = true;
          current.canFetchMore = true;
        }
        queryFn = config.fn(cursor);
      } else {
        if (current != null) {
          current.canFetchMore = false;
          stateStream.add(
            QueryState(
              snapshot: AsyncSnapshot<T>.withData(
                ConnectionState.active,
                current.data,
              ),
              updatedAt: DateTime.now().millisecondsSinceEpoch,
              failureCount: current.failureCount ?? 0,
              canFetchMore: false,
            ),
          );
        }
        return Future.value(current?.data);
      }
    } else {
      queryFn = config.fn();
    }

    if (current != null) {
      current.isFetching = true;
      stateStream.add(current);
    }

    return queryFn.then((data) {
      var next = data;
      if (config.isInfinite) {
        next = current.data ?? [];
        if (next is List) {
          next.add(data);
        } else {
          next = [data];
        }
      }

      stateStream.add(
        QueryState(
          snapshot: AsyncSnapshot<T>.withData(ConnectionState.active, next),
          updatedAt: DateTime.now().millisecondsSinceEpoch,
          failureCount: stateStream?.value?.failureCount ?? 0,
          canFetchMore: stateStream?.value?.canFetchMore,
        ),
      );
      return next;
    });
  }

  bool isStaleByTime([staleTime = 0]) {
    return (stateStream.value.isInvalidated ||
        stateStream.value.isSuccess ||
        (stateStream.value.updatedAt ?? 0) + staleTime <=
            DateTime.now().millisecondsSinceEpoch);
  }

  void invalidate() {
    stateStream.value?.isInvalidated = true;
    stateStream.add(stateStream.value);
  }

  void dispose() {
    stateStream.close();
  }
}

class QueryState<T> {
  AsyncSnapshot<T> snapshot;
  int updatedAt;
  int failureCount;
  bool isInvalidated;
  bool canFetchMore;
  bool isFetching;
  bool isFetchingMore;

  bool get isError {
    return snapshot?.hasError;
  }

  bool get isIdle {
    return snapshot?.connectionState == ConnectionState.none;
  }

  bool get isLoading {
    return snapshot?.connectionState == ConnectionState.waiting && data == null;
  }

  bool get isSuccess {
    return snapshot?.connectionState == ConnectionState.active &&
        !isError &&
        (snapshot?.hasData ?? false);
  }
  //  isPreviousData,
  //  isStale,

  T get data {
    return snapshot.data;
  }

  bool get hasData {
    return snapshot.data != null;
  }

  QueryState({
    this.snapshot = const AsyncSnapshot.nothing(),
    this.failureCount = 0,
    this.canFetchMore = false,
    this.isFetchingMore = false,
    this.isFetching = false,
    this.isInvalidated = false,
    this.updatedAt,
  });
}

class QueryConfig {
  final Function fn;
  final Function getFetchMore;
  final bool isInfinite;
  final bool enabled;
  final String queryKey;
  final Function onSuccess;
  final Function onError;
  final Function onSettled;
  final int staleTime;

  QueryConfig({
    @required this.fn,
    this.getFetchMore,
    this.queryKey,
    this.onSuccess,
    this.onError,
    this.onSettled,
    this.isInfinite = false,
    this.enabled = true,
    this.staleTime = 0,
  });
}
